@@ -283,7 +283,7 @@ class _TestGlob:
         else:
             p = [p]
         res = glob.glob(p, **kwargs)
-        print("RESULTS: ", res)
+        print("RESULTS: ", sorted(res))
         if res:
             cls.assert_equal({type(r) for r in res}, {str})
         cls.assert_count_equal(glob.iglob(p, **kwargs), res)
@@ -350,7 +350,7 @@ class _TestGlob:
         print("PATTERN: ", pattern)
         print("FLAGS: ", bin(flags))
         print("NEGATIVE: ", bin(negative))
-        print("EXPECTED: ", results)
+        print("EXPECTED: ", sorted(results) if results is not None else results)
 
         if cls.cwd_temp:
             if negative:
@@ -385,7 +385,8 @@ class Testglob(_TestGlob):
         # Glob one directory
         [('a*',), [('a',), ('aab',), ('aaa',)]],
         [('*a',), [('a',), ('aaa',)]],
-        [('.*',), [('.',), ('..',), ('.aa',), ('.bb',)]],
+        [('.*',), [('.',), ('..',), ('.aa',), ('.bb',)], glob.SCANDOTDIR],
+        [('.*',), [('.aa',), ('.bb',)]],
         [('?aa',), [('aaa',)]],
         [('aa?',), [('aaa',), ('aab',)]],
         [('aa[ab]',), [('aaa',), ('aab',)]],
@@ -913,13 +914,157 @@ class TestGlobMarked(Testglob):
     DEFAULT_FLAGS = glob.BRACE | glob.EXTGLOB | glob.GLOBSTAR | glob.FOLLOW | glob.MARK
 
 
+class TestPathlibNorm(unittest.TestCase):
+    """Test normalization cases."""
+
+    def test_norm(self):
+        """Test normalization."""
+
+        self.assertEqual(glob.Glob('.')._pathlib_norm('/./test'), '/test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('/.'), '/')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('test/.'), 'test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('test/./'), 'test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('./.'), '')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('.'), '')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('test/./.test/'), 'test/.test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('/.test/'), '/.test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('/../../././.'), '/../..')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('./././././'), '')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('././../../'), '../..')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('/././../../'), '/../..')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('/'), '/')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('/.'), '/')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('./'), '')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('./test'), 'test')
+
+    @unittest.skipUnless(sys.platform.startswith('win'), "Windows specific test")
+    def test_norm_windows(self):
+        """Test normalization on Windows."""
+
+        self.assertEqual(glob.Glob('.')._pathlib_norm('\\.\\test'), '\\test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('\\.'), '\\')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('test\\.'), 'test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('test\\.\\'), 'test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('.\\.'), '')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('.'), '')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('test\\.\\.test\\'), 'test\\.test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('\\.test\\'), '\\.test')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('\\..\\..\\.\\.\\.'), '\\..\\..')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('.\\.\\.\\.\\.\\'), '')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('.\\.\\..\\..\\'), '..\\..')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('\\.\\.\\..\\..\\'), '\\..\\..')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('\\'), '\\')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('\\.'), '\\')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('.\\'), '')
+        self.assertEqual(glob.Glob('.')._pathlib_norm('.\\test'), 'test')
+
+
 class TestHidden(_TestGlob):
     """Test hidden specific cases."""
 
     cases = [
-        [('**', '.*'), [('a', '.'), ('a', '..'), ('.aa',), ('.bb',), ('.',), ('..',)]],
-        [('*', '.*'), [('a', '.'), ('a', '..')]],
-        [('.*',), [('.aa',), ('.bb',), ('.',), ('..',)]]
+        [('**', '.*'), [('a', '.'), ('a', '..'), ('.aa',), ('.bb',), ('.',), ('..',)], glob.SCANDOTDIR],
+        [('*', '.*'), [('a', '.'), ('a', '..')], glob.SCANDOTDIR],
+        [('.*',), [('.aa',), ('.bb',), ('.',), ('..',)], glob.SCANDOTDIR],
+        [
+            ('**', '.*'),
+            [
+                ('a', '.'), ('a', '..'), ('.aa',), ('.aa', '.'), ('.aa', '..'),
+                ('.bb',), ('.bb', '.'), ('.bb', '..'), ('.',), ('..',)
+            ],
+            glob.D | glob.SCANDOTDIR
+        ],
+        [
+            ('**', '.*|**', '.', '.aa', '.'),
+            [
+                ('a', '.'), ('a', '..'), ('.aa',), ('.aa', '.'), ('.aa', '..'),
+                ('.bb',), ('.bb', '.'), ('.bb', '..'), ('.',), ('..',), ('.', '.aa', '.')
+            ],
+            glob.D | glob.S | glob.SCANDOTDIR
+        ],
+
+        # Test `pathlib` mode. `pathlib` normalizes out `.` directories, so when evaluating unique values,
+        # normalize paths with `.`.
+
+        # Prevent matching `.aa` and `.aa/.` (same with `.bb`)
+        [('**', '.*'), [('.aa',), ('.bb',)]],
+        [('**', '.*'), [('.aa',), ('.bb',)], glob.Z],
+        [('**', '.*'), [('.aa',), ('.bb',)], glob.SCANDOTDIR | glob.Z],
+
+        [
+            ('**', '.*'),
+            [
+                ('.aa',), ('.bb',)
+            ],
+            glob.D
+        ],
+        # Prevent matching `.aa/.` and `./.aa/.` as they are all the same as `.aa`
+        [
+            ('**', '.*|**', '.', '.aa', '.'),
+            [
+                ('.aa',), ('.bb',), ('.', '.aa', '.')
+            ],
+            glob.D | glob.S
+        ],
+        # Unique logic is disabled, so we can match `.aa` from one pattern and `./.aa/.` from another pattern.
+        # Duplicates are still restricted from a single pattern, so `.aa/.` is not found in the first pattern as
+        # `.aa` was already found, but unique results across multi-patterns is not enforced.
+        [
+            ('**', '.*|**', '.', '.aa', '.'),
+            [
+                ('.aa',), ('.bb',), ('.', '.aa', '.')
+            ],
+            glob.D | glob.S | glob.Q
+        ],
+        # Enable `pathlib` mode to ensure unique across multiple `pathlib` patterns.
+        [
+            ('**', '.*|**', '.', '.aa', '.'),
+            [
+                ('.aa',), ('.bb',)
+            ],
+            glob.D | glob.S | glob._PATHLIB
+        ],
+        # `NOUNIQUE` disables `pathlib` mode unique filtering.
+        [
+            ('**', '.*|**', '.', '.aa', '.'),
+            [
+                ('.aa',), ('.bb',), ('.', '.aa', '.')
+            ],
+            glob.D | glob.S | glob.Q | glob._PATHLIB
+        ],
+        # `pathlib` should still filter out duplicates if `.` and trailing slashes are normalized and
+        # a single patter is used.
+        [
+            ('**', '.*'),
+            [
+                ('.', ), ('..', ), ('.aa',), ('.bb',), ('.bb', '..'),
+                ('a', '.'), ('a', '..'), ('.aa', '..')
+            ],
+            glob.D | glob.S | glob.SCANDOTDIR | glob._PATHLIB
+        ],
+
+        Options(default_negate='**/./.*/*'),
+
+        # `SCANDOTDIR` does not change our patterns (the negate pattern for instance),
+        # just what is returned when scanning a folder with a wildcard.
+        [
+            ('**', '.*', '.aa', '*'),
+            [
+                ('.', '.bb', 'H')
+            ],
+            glob.D | glob.S | glob.N
+        ],
+
+        # Should we allow this? Or should `NODOTDIR` not apply to `NEGATE` patterns.
+        [
+            ('**', '.*', '.aa', '*'),
+            [
+                ('.', '.bb', 'H'), ('.', '.aa', 'G')
+            ],
+            glob.D | glob.S | glob.N | glob.Z
+        ],
+
+        Options(default_negate='**')
     ]
 
     @classmethod
@@ -1039,15 +1184,36 @@ class TestGlobCornerCaseMarked(Testglob):
 class TestGlobEscapes(unittest.TestCase):
     """Test escaping."""
 
-    def check_escape(self, arg, expected, raw=False, unix=None):
+    def check_escape(self, arg, expected, raw=False, unix=None, raw_chars=True):
         """Verify escapes."""
 
+        flags = 0
+        if unix is False:
+            flags = glob.FORCEWIN
+        elif unix is True:
+            flags = glob.FORCEUNIX
+
         if raw:
-            self.assertEqual(glob.raw_escape(arg, unix=unix), expected)
-            self.assertEqual(glob.raw_escape(os.fsencode(arg), unix=unix), os.fsencode(expected))
+            self.assertEqual(glob.raw_escape(arg, unix=unix, raw_chars=raw_chars), expected)
+            self.assertEqual(glob.raw_escape(os.fsencode(arg), unix=unix, raw_chars=raw_chars), os.fsencode(expected))
+            file = (util.norm_pattern(arg, False, True) if raw_chars else arg).replace('\\\\', '\\')
+            self.assertTrue(
+                glob.globmatch(
+                    file,
+                    glob.raw_escape(arg, unix=unix, raw_chars=raw_chars),
+                    flags=flags
+                )
+            )
         else:
             self.assertEqual(glob.escape(arg, unix=unix), expected)
             self.assertEqual(glob.escape(os.fsencode(arg), unix=unix), os.fsencode(expected))
+            self.assertTrue(
+                glob.globmatch(
+                    arg,
+                    glob.escape(arg, unix=unix),
+                    flags=flags
+                )
+            )
 
     def test_escape(self):
         """Test path escapes."""
@@ -1057,20 +1223,36 @@ class TestGlobEscapes(unittest.TestCase):
         check('[', r'\[')
         check('?', r'\?')
         check('*', r'\*')
-        check('[[_/*?*/_]]', r'\[\[_/\*\?\*/_]]')
-        check('/[[_/*?*/_]]/', r'/\[\[_/\*\?\*/_]]/')
+        check('[[_/*?*/_]]', r'\[\[_/\*\?\*/_\]\]')
+        check('/[[_/*?*/_]]/', r'/\[\[_/\*\?\*/_\]\]/')
 
     def test_raw_escape(self):
         """Test path escapes."""
 
         check = self.check_escape
-        check('abc', 'abc', raw=True)
-        check('[', r'\[', raw=True)
-        check('?', r'\?', raw=True)
-        check('*', r'\*', raw=True)
-        check('[[_/*?*/_]]', r'\[\[_/\*\?\*/_]]', raw=True)
-        check('/[[_/*?*/_]]/', r'/\[\[_/\*\?\*/_]]/', raw=True)
+        check(r'abc', 'abc', raw=True)
+        check(r'[', r'\[', raw=True)
+        check(r'?', r'\?', raw=True)
+        check(r'*', r'\*', raw=True)
+        check(r'[[_/*?*/_]]', r'\[\[_/\*\?\*/_\]\]', raw=True)
+        check(r'/[[_/*?*/_]]/', r'/\[\[_/\*\?\*/_\]\]/', raw=True)
         check(r'\x3f', r'\?', raw=True)
+        check(r'\\\\[^what]\\name\\temp', r'\\\\[^what]\\name\\temp', raw=True, unix=False)
+        check('//[^what]/name/temp', r'//[^what]/name/temp', raw=True, unix=False)
+
+    def test_raw_escape_no_raw_chars(self):
+        """Test path escapes with no raw character translations."""
+
+        check = self.check_escape
+        check(r'abc', 'abc', raw=True, raw_chars=False)
+        check(r'[', r'\[', raw=True, raw_chars=False)
+        check(r'?', r'\?', raw=True, raw_chars=False)
+        check(r'*', r'\*', raw=True, raw_chars=False)
+        check(r'[[_/*?*/_]]', r'\[\[_/\*\?\*/_\]\]', raw=True, raw_chars=False)
+        check(r'/[[_/*?*/_]]/', r'/\[\[_/\*\?\*/_\]\]/', raw=True, raw_chars=False)
+        check(r'\x3f', r'\\x3f', raw=True, raw_chars=False)
+        check(r'\\\\[^what]\\name\\temp', r'\\\\[^what]\\name\\temp', raw=True, raw_chars=False, unix=False)
+        check('//[^what]/name/temp', r'//[^what]/name/temp', raw=True, raw_chars=False, unix=False)
 
     @unittest.skipUnless(sys.platform.startswith('win'), "Windows specific test")
     def test_escape_windows(self):
@@ -1079,10 +1261,11 @@ class TestGlobEscapes(unittest.TestCase):
         check = self.check_escape
         check('a:\\?', r'a:\\\?')
         check('b:\\*', r'b:\\\*')
-        check(r'\\\\?\\c:\\?', r'\\\\?\\c:\\\?')
-        check(r'\\\\*\\*\\*', r'\\\\*\\*\\\*')
+        check('\\\\?\\c:\\?', r'\\\\?\\c:\\\?')
+        check('\\\\*\\*\\*', r'\\\\*\\*\\\*')
         check('//?/c:/?', r'//?/c:/\?')
         check('//*/*/*', r'//*/*/\*')
+        check('//[^what]/name/temp', r'//[^what]/name/temp')
 
     def test_escape_forced_windows(self):
         """Test forced windows escapes."""
@@ -1090,10 +1273,16 @@ class TestGlobEscapes(unittest.TestCase):
         check = self.check_escape
         check('a:\\?', r'a:\\\?', unix=False)
         check('b:\\*', r'b:\\\*', unix=False)
-        check(r'\\\\?\\c:\\?', r'\\\\?\\c:\\\?', unix=False)
-        check(r'\\\\*\\*\\*', r'\\\\*\\*\\\*', unix=False)
+        check('\\\\?\\c:\\?', r'\\\\?\\c:\\\?', unix=False)
+        check('\\\\*\\*\\*', r'\\\\*\\*\\\*', unix=False)
         check('//?/c:/?', r'//?/c:/\?', unix=False)
         check('//*/*/*', r'//*/*/\*', unix=False)
+        check(
+            '//./Volume{b75e2c83-0000-0000-0000-602f00000000}/temp',
+            r'//./Volume\{b75e2c83-0000-0000-0000-602f00000000\}/temp',
+            unix=False
+        )
+        check('//[^what]/name/temp', r'//[^what]/name/temp', unix=False)
 
     def test_escape_forced_unix(self):
         """Test forced windows Unix."""
@@ -1101,10 +1290,11 @@ class TestGlobEscapes(unittest.TestCase):
         check = self.check_escape
         check('a:\\?', r'a:\\\?', unix=True)
         check('b:\\*', r'b:\\\*', unix=True)
-        check(r'\\\\?\\c:\\?', r'\\\\\\\\\?\\\\c:\\\\\?', unix=True)
-        check(r'\\\\*\\*\\*', r'\\\\\\\\\*\\\\\*\\\\\*', unix=True)
+        check('\\\\?\\c:\\?', r'\\\\\?\\c:\\\?', unix=True)
+        check('\\\\*\\*\\*', r'\\\\\*\\\*\\\*', unix=True)
         check('//?/c:/?', r'//\?/c:/\?', unix=True)
         check('//*/*/*', r'//\*/\*/\*', unix=True)
+        check('//[^what]/name/temp', r'//\[\^what\]/name/temp', unix=True)
 
 
 @unittest.skipUnless(sys.platform.startswith('win'), "Windows specific test")
@@ -1241,6 +1431,12 @@ class TestTilde(unittest.TestCase):
         files = os.listdir(os.path.expanduser('~'))
         self.assertEqual(len(glob.glob('~/*', flags=glob.T | glob.D)), len(files))
 
+    def test_tilde_bytes(self):
+        """Test tilde in bytes."""
+
+        files = os.listdir(os.path.expanduser(b'~'))
+        self.assertEqual(len(glob.glob(b'~/*', flags=glob.T | glob.D)), len(files))
+
     def test_tilde_user(self):
         """Test tilde user cases."""
 
