@@ -2,6 +2,7 @@
 """Tests for `globmatch`."""
 import unittest
 import pytest
+import re
 import os
 import sys
 import wcmatch.glob as glob
@@ -300,6 +301,10 @@ class TestGlobFilter:
         ['a/.*/b', ['a/./b', 'a/../b', 'a/.d/b'], glob.D],
         ['a/*/b', ['a/c/b'], 0],
         ['a/.*/b', ['a/./b', 'a/../b', 'a/.d/b'], 0],
+        ['a/.*/b', ['a/.d/b'], glob.Z],
+        # Escaped `.` will still be treated as a `.`
+        [r'a/\.*/b', ['a/./b', 'a/../b', 'a/.d/b'], 0],
+        [r'a/\.*/b', ['a/.d/b'], glob.Z],
 
         # this also tests that changing the options needs
         # to change the cache key, even if the pattern is
@@ -480,7 +485,8 @@ class TestGlobFilter:
         GlobFiles(
             [
                 'a/b/.x/c', 'a/b/.x/c/d', 'a/b/.x/c/d/e', 'a/b/.x', 'a/b/.x/',
-                'a/.x/b', '.x', '.x/', '.x/a', '.x/a/b', 'a/.x/b/.x/c', '.x/.x'
+                'a/.x/b', '.x', '.x/', '.x/a', '.x/a/b', 'a/.x/b/.x/c', '.x/.x',
+                'test.x/a'
             ]
         ),
         [
@@ -491,6 +497,20 @@ class TestGlobFilter:
             ]
         ],
 
+        [
+            '**.x/*',
+            [
+                'test.x/a', '.x/a'
+            ]
+        ],
+
+        [
+            r'**\.x/*',
+            [
+                'test.x/a', '.x/a'
+            ]
+        ],
+
         # https://github.com/isaacs/minimatch/issues/59
         ['[z-a]', []],
         ['a/[2015-03-10T00:23:08.647Z]/z', []],
@@ -528,25 +548,76 @@ class TestGlobFilter:
         ['@(test\\', ['@(test\\']],
         ['@(test[)', ['test[']],
 
+        # Dot tests
+        GlobFiles(
+            [
+                '.', '..', '.abc', 'abc', '...', '..abc'
+            ]
+        ),
+
+        # Basic dot tests
+        ['[.]abc', []],
+        [r'[\.]abc', []],
+        ['.abc', ['.abc']],
+        [r'\.abc', ['.abc']],
+        ['[.]abc', ['.abc'], glob.D],
+        [r'[\.]abc', ['.abc'], glob.D],
+
+        ['.', ['.']],
+        ['..', ['..']],
+        ['.*', ['.', '..', '.abc', '...', '..abc']],
+        [r'.\a*', ['.abc']],
+        [r'\.', ['.']],
+        [r'\..', ['..']],
+        [r'\.\.', ['..']],
+        ['..*', ['..', '...', '..abc']],
+        [r'...', ['...']],
+        [r'..\.', ['...']],
+
+        ['.', ['.'], glob.Z],
+        ['..', ['..'], glob.Z],
+        ['.*', ['.abc', '...', '..abc'], glob.Z],
+        [r'.\a*', ['.abc'], glob.Z],
+        [r'\.', ['.'], glob.Z],
+        [r'\..', ['..'], glob.Z],
+        [r'\.\.', ['..'], glob.Z],
+        ['..*', ['...', '..abc'], glob.Z],
+        [r'...', ['...'], glob.Z],
+        [r'..\.', ['...'], glob.Z],
+
+        # Dot tests trailing slashes
+        GlobFiles(
+            [
+                './', '../', '.abc/', 'abc/', '.../', '..abc/'
+            ]
+        ),
+        ['./', ['./']],
+        ['../', ['../']],
+        ['..\\', [] if util.is_case_sensitive() else ['../']],
+        ['./', ['./'], glob.Z],
+        ['../', ['../'], glob.Z],
+        ['..\\', [] if util.is_case_sensitive() else ['../'], glob.Z],
+
         # Inverse dot tests
         GlobFiles(
             [
                 '.', '..', '.abc', 'abc'
             ]
         ),
+
         # We enable glob.N by default, so staring with `!`
         # is a problem without glob.M
         ['!(test)', ['abc'], glob.M],
         ['!(test)', ['.abc', 'abc'], glob.D | glob.M],
         ['.!(test)', ['.', '..', '.abc'], glob.M],
         ['.!(test)', ['.', '..', '.abc'], glob.D | glob.M],
-        ['!(.)', ['..', '.abc', 'abc'], glob.M],
-        [r'!(\.)', ['..', '.abc', 'abc'], glob.M],
-        [r'!(\x2e)', ['..', '.abc', 'abc'], glob.M | glob.R],
-        ['@(!(.))', ['..', '.abc', 'abc'], glob.M],
-        ['!(@(.))', ['..', '.abc', 'abc'], glob.M],
-        ['+(!(.))', ['..', '.abc', 'abc'], glob.M],
-        ['!(+(.))', ['.abc', 'abc'], glob.M],
+        ['!(.)', ['abc'], glob.M],
+        [r'!(\.)', ['abc'], glob.M],
+        [r'!(\x2e)', ['abc'], glob.M | glob.R],
+        ['@(!(.))', ['abc'], glob.M],
+        ['!(@(.))', ['abc'], glob.M],
+        ['+(!(.))', ['abc'], glob.M],
+        ['!(+(.))', ['abc'], glob.M],
         ['!(?)', ['abc'], glob.M],
         ['!(*)', [], glob.M],
         ['!([.])', ['abc'], glob.M],
@@ -556,10 +627,13 @@ class TestGlobFilter:
         ['@(!(.))', ['..', '.abc', 'abc'], glob.M | glob.D],
         ['!(@(.))', ['..', '.abc', 'abc'], glob.M | glob.D],
         ['+(!(.))', ['..', '.abc', 'abc'], glob.M | glob.D],
+        ['+(!(.))', ['.abc', 'abc'], glob.M | glob.D | glob.Z],
         ['!(+(.))', ['.abc', 'abc'], glob.M | glob.D],
         ['!(?)', ['.abc', 'abc'], glob.M | glob.D],
         ['!(*)', [], glob.M | glob.D],
         ['!([.])', ['.abc', 'abc'], glob.M | glob.D],
+        ['@(..|.)', ['.', '..']],
+        ['@(..|.)', ['.', '..'], glob.Z],
 
         # More extended pattern dot related tests
         ['*(.)', ['.', '..']],
@@ -612,6 +686,16 @@ class TestGlobFilter:
         ['TEST/test', [], glob.U],
         ['test\\/TEST', [], glob.U],
 
+        # Ensure we don't count slashes with `*`.
+        GlobFiles(['test/test', 'test//']),
+
+        ['test/*', ['test/test']],
+        ['test/*', ['test/test'], glob.W],
+
+        GlobFiles(['test\\test', 'test\\\\']),
+
+        ['test/*', ['test\\test'], glob.W],
+
         GlobFiles(['c:/some/path', '//host/share/some/path']),
 
         # Test Windows drive and UNC host/share case sensitivity
@@ -648,7 +732,33 @@ class TestGlobFilter:
                 b".", b"..", b"...", b'.../', b"test/", b"file", b"/file"
             ]
         ),
-        [b'**/*', [b'...', b'file', b'test/...', b'test/.file', b"/file"], glob.O | glob.D]
+        [b'**/*', [b'...', b'file', b'test/...', b'test/.file', b"/file"], glob.O | glob.D],
+
+        # Test Windows drives
+        GlobFiles(
+            [
+                '//?/UNC/LOCALHOST/c$/temp', '//./UNC/LOCALHOST/c$/temp', '//?/GLOBAL/UNC/LOCALHOST/c$/temp',
+                '//?/GLOBAL/global/UNC/LOCALHOST/c$/temp', '//?/C:/temp'
+            ]
+        ),
+
+        ['//?/unc/localhost/c$/*', ['//?/UNC/LOCALHOST/c$/temp'], glob.W],
+        ['//./unc/localhost/c$/*', ['//./UNC/LOCALHOST/c$/temp'], glob.W],
+        ['//?/global/unc/localhost/c$/*', ['//?/GLOBAL/UNC/LOCALHOST/c$/temp'], glob.W],
+        ['//?/global/global/unc/localhost/c$/*', ['//?/GLOBAL/global/UNC/LOCALHOST/c$/temp'], glob.W],
+        ['//?/c:/*', ['//?/C:/temp'], glob.W],
+
+        GlobFiles(
+            [
+                b'//?/UNC/LOCALHOST/c$/temp', b'//./UNC/LOCALHOST/c$/temp', b'//?/GLOBAL/UNC/LOCALHOST/c$/temp',
+                b'//?/GLOBAL/global/UNC/LOCALHOST/c$/temp', b'//?/C:/temp'
+            ]
+        ),
+        [b'//?/unc/localhost/c$/*', [b'//?/UNC/LOCALHOST/c$/temp'], glob.W],
+        [b'//./unc/localhost/c$/*', [b'//./UNC/LOCALHOST/c$/temp'], glob.W],
+        [b'//?/global/unc/localhost/c$/*', [b'//?/GLOBAL/UNC/LOCALHOST/c$/temp'], glob.W],
+        [b'//?/global/global/unc/localhost/c$/*', [b'//?/GLOBAL/global/UNC/LOCALHOST/c$/temp'], glob.W],
+        [b'//?/c:/*', [b'//?/C:/temp'], glob.W]
     ]
 
     @classmethod
@@ -986,8 +1096,8 @@ class TestGlobMatchSpecial(unittest.TestCase):
         if util.PY37:
             value = (
                 [
-                    '^(?s:(?:(?!(?:/|^)\\.).)*?(?:^|$|/)+'
-                    '(?!(?:\\.{1,2})(?:$|/))(?![/.])[\x00-\x7f]/+stuff/+(?=.)'
+                    '^(?s:(?:(?!(?:/|^)\\.).)*?(?:^|$|/)+' +
+                    ('(?!(?:\\.{1,2})(?:$|/))(?![/.])[\x00-\x7f]/+stuff/+(?=[^%s])' % re.escape('/')) +
                     '(?!(?:\\.{1,2})(?:$|/))(?:(?!\\.)[^/]*?)?[/]*?)$'
                 ],
                 []
@@ -995,8 +1105,8 @@ class TestGlobMatchSpecial(unittest.TestCase):
         elif util.PY36:
             value = (
                 [
-                    '^(?s:(?:(?!(?:\\/|^)\\.).)*?(?:^|$|\\/)+'
-                    '(?!(?:\\.{1,2})(?:$|\\/))(?![\\/.])[\x00-\x7f]\\/+stuff\\/+(?=.)'
+                    '^(?s:(?:(?!(?:\\/|^)\\.).)*?(?:^|$|\\/)+' +
+                    ('(?!(?:\\.{1,2})(?:$|\\/))(?![\\/.])[\x00-\x7f]\\/+stuff\\/+(?=[^%s])' % re.escape('/')) +
                     '(?!(?:\\.{1,2})(?:$|\\/))(?:(?!\\.)[^\\/]*?)?[\\/]*?)$'
                 ],
                 []
@@ -1004,8 +1114,8 @@ class TestGlobMatchSpecial(unittest.TestCase):
         else:
             value = (
                 [
-                    '(?s)^(?:(?:(?!(?:\\/|^)\\.).)*?(?:^|$|\\/)+'
-                    '(?!(?:\\.{1,2})(?:$|\\/))(?![\\/.])[\x00-\x7f]\\/+stuff\\/+(?=.)'
+                    '(?s)^(?:(?:(?!(?:\\/|^)\\.).)*?(?:^|$|\\/)+' +
+                    ('(?!(?:\\.{1,2})(?:$|\\/))(?![\\/.])[\x00-\x7f]\\/+stuff\\/+(?=[^%s])' % re.escape('/')) +
                     '(?!(?:\\.{1,2})(?:$|\\/))(?:(?!\\.)[^\\/]*?)?[\\/]*?)$'
                 ],
                 []
