@@ -34,18 +34,54 @@ BYTES = 1
 
 PATTERN_LIMIT = 1000
 
-RE_WIN_PATH = (
-    re.compile(r'((?:\\\\|/){2}[^\\/]+(?:\\\\|/){1}[^\\/]+|[a-z]:)((?:\\\\|/){1}|$)', re.I),
-    re.compile(br'((?:\\\\|/){2}[^\\/]+(?:\\\\|/){1}[^\\/]+|[a-z]:)((?:\\\\|/){1}|$)', re.I)
+RE_WIN_DRIVE_START = re.compile(r'((?:\\\\|/){2}((?:\\[^\\/]|[^\\/])+)|([\\]?[a-z][\\]?:))((?:\\\\|/)|$)', re.I)
+RE_WIN_DRIVE_LETTER = re.compile(r'([a-z]:)((?:\\|/)|$)', re.I)
+RE_WIN_DRIVE_PART = re.compile(r'((?:\\[^\\/]|[^\\/])+)((?:\\\\|/)|$)', re.I)
+RE_WIN_DRIVE_UNESCAPE = re.compile(r'\\(.)', re.I)
+
+RE_WIN_DRIVE = (
+    re.compile(
+        r'''(?x)
+        (
+            (?:\\\\|/){2}[?.](?:\\\\|/)(?:
+                [a-z]:|
+                unc(?:(?:\\\\|/)[^\\/]+){2} |
+                (?:global(?:\\\\|/))+(?:[a-z]:|unc(?:(?:\\\\|/)[^\\/]+){2}|[^\\/]+)
+            ) |
+            (?:\\\\|/){2}[^\\/]+(?:\\\\|/)[^\\/]+|
+            [a-z]:
+        )((?:\\\\|/){1}|$)
+        ''',
+        re.I
+    ),
+    re.compile(
+        br'''(?x)
+        (
+            (?:\\\\|/){2}[?.](?:\\\\|/)(?:
+                [a-z]:|
+                unc(?:(?:\\\\|/)[^\\/]+){2} |
+                (?:global(?:\\\\|/))+(?:[a-z]:|unc(?:(?:\\\\|/)[^\\/]+){2}|[^\\/]+)
+            ) |
+            (?:\\\\|/){2}[^\\/]+(?:\\\\|/)[^\\/]+|
+            [a-z]:
+        )((?:\\\\|/){1}|$)
+        ''',
+        re.I
+    )
 )
-RE_WIN_MAGIC = (
-    re.compile(r'([-!~*?(\[|^{]|(?<!\\)(?:(?:[\\]{2})*)\\(?!\\))'),
-    re.compile(br'([-!~*?(\[|^{]|(?<!\\)(?:(?:[\\]{2})*)\\(?!\\))')
+
+RE_MAGIC_ESCAPE = (
+    re.compile(r'([-!~*?()\[\]|^{}]|(?<!\\)(?:(?:[\\]{2})*)\\(?!\\))'),
+    re.compile(br'([-!~*?()\[\]|^{}]|(?<!\\)(?:(?:[\\]{2})*)\\(?!\\))')
 )
 RE_MAGIC = (
     re.compile(r'([-!~*?(\[|^{\\])'),
     re.compile(br'([-!~*?(\[|^{\\])')
 )
+RE_WIN_DRIVE_MAGIC = (
+    re.compile(r'([{}|]|(?<!\\)(?:(?:[\\]{2})*)\\(?!\\))'),
+    re.compile(br'([{}|]|(?<!\\)(?:(?:[\\]{2})*)\\(?!\\))')
+)
 RE_WIN_MOUNT = (
     re.compile(r'\\|[a-z]:(?:\\|$)', re.I),
     re.compile(br'\\|[a-z]:(?:\\|$)', re.I)
@@ -86,8 +122,8 @@ MINUS_NEGATIVE_SYM = frozenset((b'-', '-'))
 ROUND_BRACKET = frozenset((b'(', '('))
 EXT_TYPES = frozenset(('*', '?', '+', '@', '!'))
 
-# Common flags are found between `0x0001 - 0xffff`
-# Implementation specific (`glob` vs `fnmatch` vs `wcmatch`) are found between `0x00010000 - 0xffff0000`
+# Common flags are found between `0x0001 - 0xffffff`
+# Implementation specific (`glob` vs `fnmatch` vs `wcmatch`) are found between `0x01000000 - 0xff000000`
 # Internal special flags are found at `0x100000000` and above
 CASE = 0x0001
 IGNORECASE = 0x0002
@@ -109,6 +145,7 @@ FORCEWIN = 0x10000
 FORCEUNIX = 0x20000
 GLOBTILDE = 0x40000
 NOUNIQUE = 0x80000
+NODOTDIR = 0x100000
 
 # Internal flag
 _TRANSLATE = 0x100000000  # Lets us know we are performing a translation, and we just want the regex.
@@ -138,6 +175,7 @@ FLAG_MASK = (
     GLOBTILDE |
     SPLIT |
     NOUNIQUE |
+    NODOTDIR |
     _TRANSLATE |
     _ANCHOR |
     _EXTMATCHBASE |
@@ -185,6 +223,7 @@ _PATH_EOP = r'(?:$|%(sep)s)'
 # in addition to slashes.
 _GLOBSTAR_DIV = r'(?:^|$|%s)+'
 # Lookahead to see there is one character.
+_NEED_CHAR_PATH = r'(?=[^%(sep)s])'
 _NEED_CHAR = r'(?=.)'
 _NEED_SEP = r'(?=%s)'
 # Group that matches one or none
@@ -229,40 +268,113 @@ class PathNameException(Exception):
     """Path name exception."""
 
 
+class DotException(Exception):
+    """Dot exception."""
+
+
 class PatternLimitException(Exception):
     """Pattern limit exception."""
 
 
-def raw_escape(pattern, unix=None):
+def raw_escape(pattern, unix=None, raw_chars=True):
     """Apply raw character transform before applying escape."""
 
-    pattern = util.norm_pattern(pattern, False, True)
-    return escape(pattern, unix)
+    return _escape(util.norm_pattern(pattern, False, raw_chars, True), unix, True)
 
 
 def escape(pattern, unix=None):
+    """Normal escape."""
+
+    return _escape(pattern, unix)
+
+
+def _escape(pattern, unix=None, raw=False):
     """Escape."""
 
-    is_bytes = isinstance(pattern, bytes)
-    ptype = BYTES if is_bytes else UNICODE
-    replace = br'\\\1' if is_bytes else r'\\\1'
-    win = ((unix is None and util.platform() == "windows") or unix is False)
-    magic = RE_WIN_MAGIC[ptype] if win else RE_MAGIC[ptype]
+    if isinstance(pattern, bytes):
+        drive_pat = RE_WIN_DRIVE[BYTES]
+        magic = RE_MAGIC_ESCAPE[BYTES]
+        drive_magic = RE_WIN_DRIVE_MAGIC[BYTES]
+        replace = br'\\\1'
+        slash = b'\\'
+        double_slash = b'\\\\'
+        drive = b''
+    else:
+        drive_pat = RE_WIN_DRIVE[UNICODE]
+        magic = RE_MAGIC_ESCAPE[UNICODE]
+        drive_magic = RE_WIN_DRIVE_MAGIC[UNICODE]
+        replace = r'\\\1'
+        slash = '\\'
+        double_slash = '\\\\'
+        drive = ''
+
+    if not raw:
+        pattern = pattern.replace(slash, double_slash)
 
     # Handle windows drives special.
     # Windows drives are handled special internally.
     # So we shouldn't escape them as we'll just have to
     # detect and undo it later.
-    drive = b'' if is_bytes else ''
-    if win:
-        m = RE_WIN_PATH[ptype].match(pattern)
+    length = 0
+    if ((unix is None and util.platform() == "windows") or unix is False):
+        m = drive_pat.match(pattern)
         if m:
+            # Replace splitting magic chars
             drive = m.group(0)
-    pattern = pattern[len(drive):]
+            length = len(drive)
+            drive = drive_magic.sub(replace, m.group(0))
+    pattern = pattern[length:]
 
     return drive + magic.sub(replace, pattern)
 
 
+def _get_win_drive(pattern, regex=False, case_sensitive=False):
+    """Get Windows drive."""
+
+    drive = None
+    slash = False
+    end = 0
+    root_specified = False
+    m = RE_WIN_DRIVE_START.match(pattern)
+    if m:
+        end = m.end(0)
+        if m.group(3) and RE_WIN_DRIVE_LETTER.match(m.group(0)):
+            if regex:
+                drive = escape_drive(RE_WIN_DRIVE_UNESCAPE.sub(r'\1', m.group(3)), case_sensitive)
+            else:
+                drive = RE_WIN_DRIVE_UNESCAPE.sub(r'\1', m.group(0))
+            slash = bool(m.group(4))
+            root_specified = True
+        elif m.group(2):
+            root_specified = True
+            part = [RE_WIN_DRIVE_UNESCAPE.sub(r'\1', m.group(2))]
+            is_special = part[-1].lower() in ('.', '?')
+            complete = 1
+            first = 1
+            count = 0
+            for count, m in enumerate(RE_WIN_DRIVE_PART.finditer(pattern, m.end(0)), 1):
+                end = m.end(0)
+                part.append(RE_WIN_DRIVE_UNESCAPE.sub(r'\1', m.group(1)))
+                slash = bool(m.group(2))
+                if is_special:
+                    if count == first and part[-1].lower() == 'unc':
+                        complete += 2
+                    elif count == first and part[-1].lower() == 'global':
+                        first += 1
+                        complete += 1
+                if count == complete:
+                    break
+            if count == complete:
+                if not regex:
+                    drive = '\\\\{}{}'.format('\\'.join(part), '\\' if slash else '')
+                else:
+                    drive = r'[\\/]{2}' + r'[\\/]'.join([escape_drive(p, case_sensitive) for p in part])
+    elif pattern.startswith('\\\\'):
+        root_specified = True
+
+    return root_specified, drive, slash, end
+
+
 def is_negative(pattern, flags):
     """Check if negative pattern."""
 
@@ -655,15 +767,14 @@ class WcPathSplit(object):
 
         # Detect and store away windows drive as a literal
         if self.win_drive_detect:
-            m = RE_WIN_PATH[UNICODE].match(pattern)
-            if m:
-                drive = m.group(0).replace('\\\\', '\\')
+            root_specified, drive, slash, end = _get_win_drive(pattern)
+            if drive is not None:
                 if self.is_bytes:
                     drive = drive.encode('latin-1')
                 parts.append(WcGlob(drive, False, False, True, True))
-                start = m.end(0) - 1
-                i.advance(start + 1)
-            elif pattern.startswith('\\\\'):
+                start = end - 1
+                i.advance(start)
+            elif drive is None and root_specified:
                 parts.append(WcGlob(b'\\' if self.is_bytes else '\\', False, False, True, True))
                 start = 1
                 i.advance(2)
@@ -877,6 +988,7 @@ class WcParse(object):
         self.extmatchbase = bool(flags & _EXTMATCHBASE)
         self.rtl = bool(flags & _RTL)
         self.anchor = bool(flags & _ANCHOR)
+        self.nodotdir = bool(flags & NODOTDIR)
         self.case_sensitive = get_case(flags)
         self.in_list = False
         self.flags = flags
@@ -902,6 +1014,10 @@ class WcParse(object):
         self.path_star_dot2 = _PATH_STAR_NO_DOTMATCH % sep
         self.path_gstar_dot1 = _PATH_GSTAR_DOTMATCH % sep
         self.path_gstar_dot2 = _PATH_GSTAR_NO_DOTMATCH % sep
+        if self.pathname:
+            self.need_char = _NEED_CHAR_PATH % sep
+        else:
+            self.need_char = _NEED_CHAR
 
     def set_after_start(self):
         """Set tracker for character after the start of a directory."""
@@ -1052,6 +1168,8 @@ class WcParse(object):
                 subindex = i.index
                 try:
                     value = self._references(i, True)
+                except DotException:
+                    value = re.escape(next(i))
                 except PathNameException:
                     raise StopIteration
                 except StopIteration:
@@ -1125,15 +1243,74 @@ class WcParse(object):
                 i.rewind(1)
             else:
                 value = re.escape(c)
+        elif c == '.':
+            # Let dots be handled special
+            i.rewind(1)
+            raise DotException
         else:
             # \a, \b, \c, etc.
             value = re.escape(c)
-            if c == '.' and self.after_start and self.in_list:
-                self.allow_special_dir = True
-                self.reset_dir_track()
 
         return value
 
+    def _handle_dot(self, i, current):
+        """Handle dot."""
+
+        is_current = True
+        is_previous = False
+
+        if self.after_start and self.pathname and self.nodotdir:
+            try:
+                index = i.index
+                while True:
+                    c = next(i)
+                    if c == '.' and is_current:
+                        is_previous = True
+                        is_current = False
+                    elif c == '.' and is_previous:
+                        is_previous = False
+                        raise StopIteration
+                    elif c in ('|', ')') and self.in_list:
+                        raise StopIteration
+                    elif c == '\\':
+                        try:
+                            self._references(i, True)
+                            # Was not what we expected
+                            is_current = False
+                            is_previous = False
+                            raise StopIteration
+                        except DotException:
+                            if is_current:
+                                is_previous = True
+                                is_current = False
+                                c = next(i)
+                            else:
+                                is_previous = False
+                                raise StopIteration
+                        except PathNameException:
+                            # Looks like escape was a valid slash
+                            # Store pattern accordingly
+                            raise StopIteration
+                        except StopIteration:
+                            # Ran out of characters so assume backslash
+                            if self.sep != '\\' or not self.pathname:
+                                is_current = False
+                                is_previous = False
+                            raise StopIteration
+                    elif c == '/':
+                        raise StopIteration
+                    else:
+                        is_current = False
+                        is_previous = False
+                        raise StopIteration
+            except StopIteration:
+                i.rewind(i.index - index)
+
+        if not is_current and not is_previous:
+            current.append(r'(?!\.[.]?{})\.'.format(self.path_eop))
+        else:
+            current.append(re.escape('.'))
+
     def _handle_star(self, i, current):
         """Handle star."""
 
@@ -1177,6 +1354,8 @@ class WcParse(object):
                             self._references(i, True)
                             # Was not what we expected
                             # Assume two single stars
+                        except DotException:
+                            pass
                         except PathNameException:
                             # Looks like escape was a valid slash
                             # Store pattern accordingly
@@ -1200,7 +1379,7 @@ class WcParse(object):
                     value = globstar
 
         if self.after_start and value != globstar:
-            value = _NEED_CHAR + value
+            value = self.need_char + value
             # Consume duplicate starts
             try:
                 c = next(i)
@@ -1265,8 +1444,9 @@ class WcParse(object):
         temp_in_list = self.in_list
         temp_inv_ext = self.inv_ext
         self.in_list = True
+
         if reset_dot:
-            self.allow_special_dir = False
+            self.match_dot_dir = False
 
         # Start list parsing
         success = True
@@ -1285,10 +1465,11 @@ class WcParse(object):
                     pass
                 elif c == '*':
                     self._handle_star(i, extended)
-                elif c == '.' and self.after_start:
-                    extended.append(re.escape(c))
-                    self.allow_special_dir = True
-                    self.reset_dir_track()
+                elif c == '.':
+                    self._handle_dot(i, extended)
+                    if self.after_start:
+                        self.match_dot_dir = self.dot and not self.nodotdir
+                        self.reset_dir_track()
                 elif c == '?':
                     extended.append(self._restrict_sequence() + _QMARK)
                 elif c == '/':
@@ -1303,6 +1484,8 @@ class WcParse(object):
                 elif c == '\\':
                     try:
                         extended.append(self._references(i))
+                    except DotException:
+                        continue
                     except StopIteration:
                         # We've reached the end.
                         # Do nothing because this is going to abort the `extmatch` anyways.
@@ -1333,7 +1516,7 @@ class WcParse(object):
                 # If pattern is at the end, anchor the match to the end.
                 current.append(_EXCLA_GROUP % ''.join(extended))
                 if self.pathname:
-                    if not temp_after_start or self.allow_special_dir:
+                    if not temp_after_start or self.match_dot_dir:
                         star = self.path_star
                     elif temp_after_start and not self.dot:
                         star = self.path_star_dot2
@@ -1346,7 +1529,7 @@ class WcParse(object):
                         star = _NO_DOT + _STAR
 
                 if temp_after_start:
-                    star = _NEED_CHAR + star
+                    star = self.need_char + star
                 # Place holder for closing, but store the proper star
                 # so we know which one to use
                 current.append(InvPlaceholder(star))
@@ -1402,19 +1585,14 @@ class WcParse(object):
         iter(i)
         root_specified = False
         if self.win_drive_detect:
-            m = RE_WIN_PATH[UNICODE].match(pattern)
-            if m:
-                drive = m.group(0).replace('\\\\', '\\')
-                if drive.endswith('\\'):
-                    slash = True
-                drive = drive[:-1]
-                current.append(escape_drive(drive, self.case_sensitive))
+            root_specified, drive, slash, end = _get_win_drive(pattern, True, self.case_sensitive)
+            if drive is not None:
+                current.append(drive)
                 if slash:
                     current.append(self.get_path_sep() + _ONE_OR_MORE)
-                i.advance(m.end(0))
+                i.advance(end)
                 self.consume_path_sep(i)
-                root_specified = True
-            elif pattern.startswith('\\\\'):
+            elif drive is None and root_specified:
                 root_specified = True
         elif not self.win_drive_detect and self.pathname and pattern.startswith('/'):
             root_specified = True
@@ -1437,6 +1615,8 @@ class WcParse(object):
             if self.extend and c in EXT_TYPES and self.parse_extend(c, i, current, True):
                 # Nothing to do
                 pass
+            elif c == '.':
+                self._handle_dot(i, current)
             elif c == '*':
                 self._handle_star(i, current)
             elif c == '?':
@@ -1459,6 +1639,8 @@ class WcParse(object):
                         self.consume_path_sep(i)
                         self.matchbase = False
                     current.append(value)
+                except DotException:
+                    continue
                 except StopIteration:
                     i.rewind(i.index - index)
                     current.append(re.escape(c))
@@ -1495,6 +1677,7 @@ class WcParse(object):
         p = p.decode('latin-1') if self.is_bytes else p
 
         if self.negative:
+            # TODO: Do we prevent `NODOTDIR` for negative patterns?
             self.globstar_capture = False
             self.dot = True
 
@@ -1512,16 +1695,22 @@ class WcParse(object):
             self.globstar = globstar
 
         elif self.rtl:
+            # Add a `**` that can capture anything: dots, special directories, symlinks, etc.
+            # We are simulating right to left, so everything on the left should be accepted without
+            # question.
             globstar = self.globstar
             dot = self.dot
             gstar = self.path_gstar_dot1
+            globstar_capture = self.globstar_capture
             self.path_gstar_dot1 = _PATH_GSTAR_RTL_MATCH
             self.dot = True
             self.globstar = True
+            self.globstar_capture = False
             self.root('**', prepend)
             self.globstar = globstar
             self.dot = dot
             self.path_gstar_dot1 = gstar
+            self.globstar_capture = globstar_capture
 
         if p:
             self.root(p, result)
